#### Generate measured time points ####
## solve for t_ikj
time.measured <- function(i, k, integral.lambda_kt){
  t_ik <- 0
  u_ikj <- runif(1)
  t_ikj <- -log(u_ikj)/(exp(Z_i[i,-1] %*% gamma_k[,k])*integral.lambda_kt)
  # print(u_ik1)
  while(t_ikj < T_i[i]){
    t_ik <- append(t_ik, t_ikj)
    u_ikj <- runif(1)
    t_ikj <- -log(u_ikj)/(exp(Z_i[i,-1] %*% gamma_k[,k])*integral.lambda_kt) + t_ikj
    # print(t_ikj)
  }
  return(t_ik)
}

## Calculate Sigma2(t) using measured time points
timepoints.cov.coef <- function(timepoints, b){
  sapply(timepoints, function(t) {exp(-((t-timepoints)/b)^2)})
}

#### Find closest responses at time t ####
## find measured values X_ikj whose measured timepoint t_ikj is the closest to t for k-th process of all patients
closest_Obs <- function(obsdf, ttt, kkk){
  out <- obsdf %>% 
    filter(k==kkk) %>% 
    group_by(i) %>% 
    summarize(value_at_closest_t=value[which.min(abs(time-ttt))]) %>%
    dplyr::select(value_at_closest_t) %>%
    unlist %>% unname
  return(out)
}

#### Gauss-Hermite quadratures ####
hermite <- function (points, z) {
  p1 <- 1/pi^0.4
  p2 <- 0
  for (j in 1:points) {
    p3 <- p2
    p2 <- p1
    p1 <- z * sqrt(2/j) * p2 - sqrt((j - 1)/j) * p3
  }
  pp <- sqrt(2 * points) * p2
  c(p1, pp)
}

gauss.hermite <- function (points, iterlim = 50) {
  x <- w <- rep(0, points)
  m <- (points + 1)/2
  for (i in 1:m) {
    z <- if (i == 1) 
      sqrt(2 * points + 1) - 2 * (2 * points + 1)^(-1/6)
    else if (i == 2) 
      z - sqrt(points)/z
    else if (i == 3 || i == 4) 
      1.9 * z - 0.9 * x[i - 2]
    else 2 * z - x[i - 2]
    for (j in 1:iterlim) {
      z1 <- z
      p <- hermite(points, z)
      z <- z1 - p[1]/p[2]
      if (abs(z - z1) <= 1e-15) 
        break
    }
    if (j == iterlim) 
      warning("iteration limit exceeded")
    x[points + 1 - i] <- -(x[i] <- z)
    w[i] <- w[points + 1 - i] <- 2/p[2]^2
  }
  r <- data.frame(points=x * sqrt(2), weights=w/sum(w))
  return(r)
}

bigauss.hermite <- function(n, mu=c(0,0), sigma=diag(2), prune=NULL) {
  if(!all(dim(sigma) == length(mu)))
    stop("mu and sigma have nonconformable dimensions")
  
  dm  <- length(mu)
  gh  <- gauss.hermite(n)
  #idx grows exponentially in n and dm
  idx <- as.matrix(expand.grid(rep(list(1:n),dm)))
  pts <- matrix(gh[idx,1],nrow(idx),dm)
  wts <- apply(matrix(gh[idx,2],nrow(idx),dm), 1, prod)
  
  ## prune
  if(!is.null(prune)) {
    qwt <- quantile(wts, probs=prune)
    pts <- pts[wts > qwt,]
    wts <- wts[wts > qwt]
  }
  return(list(points=pts, weights=wts))
}

#### Square root of a covariance maxtrix ####
mat.sqroot.2dsym <- function(a,d,b){
  tau <- a+d
  delta <- sqrt(a*d-b^2)
  R.11 <- (a+delta)/sqrt(tau+2*delta)
  R.12 <- R.21 <- b/sqrt(tau+2*delta)
  R.22 <- (d+delta)/sqrt(tau+2*delta)
  
  return(matrix(c(R.11,R.12,R.21,R.22),nrow=2,byrow=T))
}

#### Partial R.kl partial sigma.sq.k, sigma.sq.l, sigma.kl ####
partial.R.kl <- function(a,d,b){
  tau <- a+d
  delta <- sqrt(a*d-b^2)
  
  p.sigma.sq.k.11 <- ((1+1/2*d/delta)*sqrt(tau+2*delta) - (a+delta)/2/sqrt(tau+2*delta)*(1+d/delta))/(tau+2*delta)
  p.sigma.sq.k.21 <- p.sigma.sq.k.12 <- b*(-1/2)*((tau+2*delta)^(-3/2))*(1+d/delta)
  p.sigma.sq.k.22 <- ((1/2*d/delta)*sqrt(tau+2*delta) - (d+delta)/2/sqrt(tau+2*delta)*(1+d/delta))/(tau+2*delta)
  
  p.sigma.sq.l.11 <- ((1/2*a/delta)*sqrt(tau+2*delta) - (a+delta)/2/sqrt(tau+2*delta)*(1+a/delta))/(tau+2*delta)
  p.sigma.sq.l.21 <- p.sigma.sq.l.12 <- b*(-1/2)*((tau+2*delta)^(-3/2))*(1+a/delta)
  p.sigma.sq.l.22 <- ((1+1/2*a/delta)*sqrt(tau+2*delta) - (d+delta)/2/sqrt(tau+2*delta)*(1+a/delta))/(tau+2*delta)
  
  p.sigma.kl.11 <- ((-b/delta)*sqrt(tau+2*delta) - (a+delta)/sqrt(tau+2*delta)*(-b/delta))/(tau+2*delta)
  p.sigma.kl.21 <- p.sigma.kl.12 <- (sqrt(tau+2*delta)-b/sqrt(tau+2*delta)*(-b/delta))/(tau+2*delta)
  p.sigma.kl.22 <- ((-b/delta)*sqrt(tau+2*delta) - (d+delta)/sqrt(tau+2*delta)*(-b/delta))/(tau+2*delta)
  
  return(list(p.sigma.sq.k=matrix(c(p.sigma.sq.k.11,p.sigma.sq.k.12,p.sigma.sq.k.21,p.sigma.sq.k.22),nrow=2,byrow=T),
              p.sigma.sq.l=matrix(c(p.sigma.sq.l.11,p.sigma.sq.l.12,p.sigma.sq.l.21,p.sigma.sq.l.22),nrow=2,byrow=T),
              p.sigma.kl=matrix(c(p.sigma.kl.11,p.sigma.kl.12,p.sigma.kl.21,p.sigma.kl.22),nrow=2,byrow=T)
  ))
} # partial derivatives for elements in the square root of a covariance maxtrix


## E.ik, l.k and their partial derivatives for beta.k ####
E.ik.and.derivs <- function(Z.i, beta.k, gfun.inv.k, gfun.inv.k.deriv,
                            n, mu.k=0, sigma.sq.k, prune.fraction=NULL, var.deriv=FALSE, gh){
  u.k <- as.numeric(t(beta.k) %*% Z.i) + gh$points*sqrt(sigma.sq.k) + mu.k
  
    return(list(val = sum(gfun.inv.k(u.k) * gh$weights)
                , beta.k.deriv = sum(gfun.inv.k.deriv(u.k) * gh$weights) * t(Z.i)
    ))
} # E[Y_ik(t)|X_i] in equation (5) in the manuscript and its gradient

l.k.and.derivs.v2 <- function(k, K.an=kernel.func, an, t, Z, t_X_k,
                           beta.k, gfun.inv.k, gfun.inv.k.deriv,
                           n, mu.k=0, sigma.sq.k, prune.fraction=NULL,
                           gamma.hat.k, theta.hat.k=NULL, L.k=NULL, var.deriv=FALSE, gh){
  l.k <- matrix(0, ncol(Z)) # a (m+1)*1 matrix
  l.k.beta.k.deriv <- matrix(0, ncol(Z), ncol(Z))
  
  # beta.k: (m+1)*1 vector
  beta.k <- as.matrix(beta.k)
  # gamma.hat.k: m*1 vector
  gamma.hat.k <- as.matrix(gamma.hat.k)
  
  # Z: N*(m+1) matrix, the row vector is (1, age, sex) for the ith patient 
  for (i in unique(t_X_k[,1])){
    # Z.i: (m+1)*1 vector
    Z.i <- as.matrix(Z[i,])
    E.ik.results <- E.ik.and.derivs(Z.i, beta.k, gfun.inv.k, gfun.inv.k.deriv, 
                                    n, mu.k, sigma.sq.k, prune.fraction, var.deriv, gh)
    t_X_ik <- matrix(t_X_k[t_X_k[,1]==i,], ncol=5)  
    # t.k: a list containing all measured timepoints of the kth temporal process for the all patients 
    for (j in 1:nrow(t_X_ik)){
      t.ikj <- t_X_ik[j,4]
      # X.k: a list containing all measured values of the kth temporal process for the all patients, each element is a vector
      X.ik_t.ikj <- t_X_ik[j,5]
      dN.ik_t.ikj <- exp(as.numeric(-t(gamma.hat.k) %*% Z.i[-1,]))
      
      # return a (m+1)*1 matrix
      l.k <- l.k + K.an(an, t.ikj, t) * Z.i * (X.ik_t.ikj - E.ik.results$val) * dN.ik_t.ikj 
      
      # return a (m+1)*(m+1) matrix
      l.k.beta.k.deriv <- l.k.beta.k.deriv + K.an(an, t.ikj, t) * Z.i %*% (- E.ik.results$beta.k.deriv) * dN.ik_t.ikj 
    }
  }
  return(list(val=l.k, beta.k.deriv=l.k.beta.k.deriv))
} # LHS of equation (5) in the manuscript and its gradient

#### E.ikl, l.kl and their partial derivatives for beta.k, beta.l, sigma.sq.k, sigma.sq.l, sigma.kl ####
E.ikl.and.derivs <- function(Z.i, beta.1, gfun.inv.1, gfun.inv.1.deriv, 
                             beta.2, gfun.inv.2, gfun.inv.2.deriv,
                             n=10, mu.12=c(0,0), sigma.sq.1, sigma.sq.2, sigma.12, prune.fraction=NULL,
                             var.deriv.1=FALSE, var.deriv.2=FALSE, gh){
  # gh <- bigauss.hermite(n) # bivariate Gauss Hermite points and weights
  
  R.kl <- mat.sqroot.2dsym(sigma.sq.1, sigma.sq.2, sigma.12) # a 2x2 matrix
  u.k <- as.numeric(t(beta.1) %*% Z.i) + mu.12[1] + R.kl[1,1]*gh$points[,1] + R.kl[2,1]*gh$points[,2]
  u.l <- as.numeric(t(beta.2) %*% Z.i) + mu.12[2] + R.kl[1,2]*gh$points[,1] + R.kl[2,2]*gh$points[,2]
  
  p.R.kl <- partial.R.kl(sigma.sq.1, sigma.sq.2, sigma.12)
  p.R.kl.sigma.sq.k <- p.R.kl$p.sigma.sq.k
  p.R.kl.sigma.sq.l <- p.R.kl$p.sigma.sq.l
  
  p.R.kl.sigma.kl <- p.R.kl$p.sigma.kl
  partial.u.k.sigma.kl <- p.R.kl.sigma.kl[1,1]*gh$points[,1] + p.R.kl.sigma.kl[2,1]*gh$points[,2]
  partial.u.l.sigma.kl <- p.R.kl.sigma.kl[1,2]*gh$points[,1] + p.R.kl.sigma.kl[2,2]*gh$points[,2]
  
  rrr <- list(val = sum(gfun.inv.1(u.k) * gfun.inv.2(u.l) * gh$weights),
              sigma.kl.deriv = sum((gfun.inv.1.deriv(u.k) * partial.u.k.sigma.kl * gfun.inv.2(u.l) 
                                    + gfun.inv.1(u.k) * gfun.inv.2.deriv(u.l) * partial.u.l.sigma.kl) * gh$weights)
  )
  
  return(rrr)
} # E[Y_ik(t)Y_il(t)|X_i] in equation (6) in the manuscript and its gradient


l.kl.and.derivs.v2 <- function(k1, k2, K.an=kernel.func, an, t, Z, t_X_1, t_X_2, 
                            beta.1, gfun.inv.1, gfun.inv.1.deriv, beta.2, gfun.inv.2, gfun.inv.2.deriv,
                            n=10, mu.12=c(0,0), sigma.sq.1, sigma.sq.2, sigma.12, prune.fraction=NULL,
                            gamma.hat.1, gamma.hat.2,
                            var.deriv.1=FALSE, var.deriv.2=FALSE, gh, ID_match){
  l.kl <- 0
  l.kl.sigma.kl.deriv <- 0
  
  # beta.k: (m+1)*1 vector
  beta.1 <- as.matrix(beta.1)
  beta.2 <- as.matrix(beta.2)
  # gamma.hat.k: m*1 vector
  gamma.hat.1 <- as.matrix(gamma.hat.1)
  gamma.hat.2 <- as.matrix(gamma.hat.2)
  
    # Z: N*(m+1) matrix, the row vector is (1, age, sex) for the ith patient 
    for (i in 1:nrow(Z)){
      Z.i <- as.matrix(Z[i,])
      E.ikl.results <- E.ikl.and.derivs(Z.i, beta.1, gfun.inv.1, gfun.inv.1.deriv, 
                                        beta.2, gfun.inv.2, gfun.inv.2.deriv,
                                        n, mu.12, sigma.sq.1, sigma.sq.2, sigma.12, prune.fraction=NULL,
                                        var.deriv.1, var.deriv.2, gh)
      t_X_ik <- matrix(t_X_1[t_X_1[,1]==ID_match[i,1],], ncol=5)
      t_X_il <- matrix(t_X_2[t_X_2[,1]==ID_match[i,2],], ncol=5)
      
      # t.k: a list containing all measured timepoints of the kth temporal process for the all patients 
      for (j in 1:nrow(t_X_ik)){
        t.ikj <- t_X_ik[j,4]
        # X.k: a list containing all measured values of the kth temporal process for the all patients
        X.ik_t.ikj <- t_X_ik[j,5]
        dN.ik_t.ikj <- exp(as.numeric(-t(gamma.hat.1) %*% Z.i[-1,]))
        
        for (jj in 1:nrow(t_X_il)){
          t.iljj <- t_X_il[jj,4]
          X.il_t.iljj <- t_X_il[jj,5]
          dN.il_t.iljj <- exp(as.numeric(-t(gamma.hat.2) %*% Z.i[-1,]))
          
          # return a scalar
          l.kl <- l.kl + K.an(an, t.ikj, t) * K.an(an, t.iljj, t)  * (X.ik_t.ikj * X.il_t.iljj - E.ikl.results$val) * dN.ik_t.ikj * dN.il_t.iljj
          
          # return a scalar
          l.kl.sigma.kl.deriv <- l.kl.sigma.kl.deriv + K.an(an, t.ikj, t) * K.an(an, t.iljj, t)  * (- E.ikl.results$sigma.kl.deriv) * dN.ik_t.ikj * dN.il_t.iljj
        }
      }
    }
    return(list(val=l.kl, 
                sigma.kl.deriv=l.kl.sigma.kl.deriv
    ))
} # LHS of equation (6) in the manuscript and its gradient



#### Objective and gradient functions ####
obj.func.k.v3 <- function(parmt, kernel.func.k, an.k, t, Z, gh, bigh, k, time_Xikj,
                          sigma.sq.k, gfun.inv.k, gfun.inv.k.deriv, gamma.hat.k, var.deriv.k){

  l_k <- l.k.and.derivs.v2(k, kernel.func.k, an.k, t, Z, t_X_k=time_Xikj,
                           beta.k=parmt, gfun.inv.k, gfun.inv.k.deriv,
                           n, mu.k=0, sigma.sq.k, prune.fraction=NULL,
                           gamma.hat.k, theta.hat.k=NULL, L.k=NULL, var.deriv.k, gh)$val/N
  
  W <- as.numeric(t(l_k) %*% l_k)  # objective value
  
  return(W)
} # objective function for solving equation (5) in the manuscript

grad.func.k.v3 <- function(parmt, kernel.func.k, an.k, t, Z, gh, bigh, k, time_Xikj,
                           sigma.sq.k, gfun.inv.k, gfun.inv.k.deriv, gamma.hat.k, var.deriv.k){

  l.k.results <- l.k.and.derivs.v2(k, kernel.func.k, an.k, t, Z, t_X_k=time_Xikj, 
                                   parmt, gfun.inv.k, gfun.inv.k.deriv,
                                   n, mu.k=0, sigma.sq.k, prune.fraction=NULL,
                                   gamma.hat.k, theta.hat.k=NULL, L.k=NULL, var.deriv.k, gh)
  
  par.beta.grad <- as.vector(2*(t(l.k.results$val/N) %*% l.k.results$beta.k.deriv/N))
  
  return(par.beta.grad)
} # gradient function for solving equation (5) in the manuscript

obj.func.kl.v3 <- function(parmt, k, l, kernel.func.kl, an.kl, t, Z, t_X_1=time_Xikj, t_X_2=time_Xilj,
                           beta.k, gfun.inv.k, gfun.inv.k.deriv, 
                           beta.l, gfun.inv.l, gfun.inv.l.deriv,
                           n=10, mu.12=c(0,0), sigma.sq.k, sigma.sq.l, prune.fraction=NULL,
                           gamma.hat.k, gamma.hat.l,
                           var.deriv.k, var.deriv.l, gh, bigh, ID_match){
  
  l_kl <- l.kl.and.derivs.v2(k, l, kernel.func.kl, an.kl, t, Z, t_X_1, t_X_2,
                             beta.k, gfun.inv.k, gfun.inv.k.deriv, beta.l, gfun.inv.l, gfun.inv.l.deriv,
                             n=10, mu.12=c(0,0), sigma.sq.k, sigma.sq.l, parmt, prune.fraction=NULL,
                             gamma.hat.k, gamma.hat.l,
                             var.deriv.k, var.deriv.l, bigh, ID_match)$val/N
  W <- l_kl^2*N^2
  return(W)
} # objective function for solving equation (6) in the manuscript

grad.func.kl.v3 <- function(parmt, k, l, kernel.func.kl, an.kl, t, Z, t_X_1=time_Xikj, t_X_2=time_Xilj,
                            beta.k, gfun.inv.k, gfun.inv.k.deriv, beta.l, gfun.inv.l, gfun.inv.l.deriv,
                            n=10, mu.12=c(0,0), sigma.sq.k, sigma.sq.l, prune.fraction=NULL,
                            gamma.hat.k, gamma.hat.l,
                            var.deriv.k, var.deriv.l, gh, bigh, ID_match){
  
  l.kl.results <- l.kl.and.derivs.v2(k, l, kernel.func.kl, an.kl, t, Z, t_X_1, t_X_2, 
                                     beta.k, gfun.inv.k, gfun.inv.k.deriv, beta.l, gfun.inv.l, gfun.inv.l.deriv, 
                                     n=10, mu.12=c(0,0), sigma.sq.k, sigma.sq.l, parmt, prune.fraction=NULL,
                                     gamma.hat.k, gamma.hat.l,
                                     var.deriv.k, var.deriv.l, bigh, ID_match)
  
  par.cov.grad <- 2*l.kl.results$val * l.kl.results$sigma.kl.deriv
  
  return(par.cov.grad)  
} # gradient function for solving equation (6) in the manuscript